#!/usr/bin/env python3

"""
File functions.
"""

import contextlib
import io
import logging
import pathlib
import shutil
import time
from typing import Any


LOGGER = logging.getLogger(__name__)


def backup_copy(path: pathlib.Path) -> None:
    """
    Attempt to backup a copy to a timestamped path in the same directory.

    Args:
        path:
            The path to back up. If it does not exist then nothing will be done.
    """
    if path.exists():
        while True:
            timestamp = time.strftime("%Y_%m_%d_%H_%M_%S")
            new_suffix = f".{timestamp}{path.suffix}"
            backup_path = path.with_suffix(new_suffix)
            if not backup_path.exists():
                break
        LOGGER.info("Backing up %s to %s", path, backup_path)
        if path.is_dir():
            shutil.copytree(path, backup_path)
        else:
            shutil.copy2(path, backup_path)


@contextlib.contextmanager
def safe_open(
    path: pathlib.Path, *args: Any, backup: bool = False, **kwargs: Any
) -> io.TextIOBase:
    """
    Context manager for safely opening a path by first creating any missing
    parent directories and then moving any existing file to a backup path.

    Args:
        path:
            The path to open.

        backup:
            If True, backup the file at the path if it exists.

        *args:
            Positional arguments passed through to pathlib.Path.open.

        **kwargs:
            Keyword arguments passed through to pathlib.Path.open.
    Returns:
        The open file handle.
    """
    path = pathlib.Path(path).resolve()
    LOGGER.debug("opening %s", path)

    path.parent.mkdir(parents=True, exist_ok=True)

    if backup:
        backup_copy(path)

    kwargs.setdefault("encoding", "utf-8")
    with path.open(*args, **kwargs) as handle:  # pylint: disable=unspecified-encoding
        yield handle



#!/usr/bin/env python3
"""
Type definitions and functions.
"""

import logging
import numbers
import pathlib
from collections.abc import Generator, Iterable
from typing import Any, Callable, Tuple, Union

from omegaconf import DictConfig


LOGGER = logging.getLogger(__name__)


# Generic number type.
Number = Union[int, float]

# Optimization value for the Optuna sweeper returned by hydronaut.run.Runner.__call__.
OptimizationValue = Union[Number, Tuple[Number, ...]]

# Generic path type for functions that can handle both pathlib Paths or strings.
Path = Union[str, pathlib.Path]

# A type supported by the decorator.
Decorable = Union[
    Callable[DictConfig, OptimizationValue],
    Callable[DictConfig, Callable[[], OptimizationValue]],
]

# The return type of the decorator.
Runable = Callable[[], OptimizationValue]


def ensure_numeric(value: Any) -> Generator[numbers.Real]:
    """
    Convert values to numeric values. If the value is iterable, then the
    function will be recursively applied to its elements.

    Args:
        value:
            The value to check.

    Returns:
        A generator over numeric values (instances of numbers.Real). Values that
        could not be converted will be returned as NaN values.
    """
    if isinstance(value, Iterable):
        for val in value:
            yield from ensure_numeric(val)
        return
    if isinstance(value, numbers.Real):
        yield value
        return
    try:
        yield float(value)
    except ValueError:
        LOGGER.warning("Failed to convert %s to numeric value.", value)
        yield float("nan")



#!/usr/bin/env python3
"""
PyTorch Lightning experiments.
"""

import logging

import mlflow.pytorch
import lightning
from lightning.pytorch.callbacks import Callback
from omegaconf import DictConfig

from hydronaut.experiment import Experiment
from hydronaut.hydra.omegaconf import get_container
from hydronaut.types import OptimizationValue

LOGGER = logging.getLogger(__name__)


class LightningExperiment(Experiment):
    """
    Experiment class for PyTorch Lightning experiments.

    The configuration file should include a dictionary under
    "experiment.params.trainer" that contains PyTorch Lighting Trainer keyword
    arguments and their values. These will be passed through to the Trainer. See
    https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api
    for details.
    """

    def __init__(
        self,
        config: DictConfig,
        model_cls: lightning.LightningModule,
        data_module_cls: lightning.LightningDataModule,
        callback_classes: list[Callback] = None,
    ) -> None:
        """
        Args:
            config:
                Same as Experiment.__init__().

            model_cls:
                A subclass of LightningModule.

            data_module_cls:
                A subclass of LightningDataModule.

            callback_classes:
                An optional list of Callback subclasses.

        model_cls, data_module_cls and all classes in callback_classes should
        all accept a single configuration object as a parameter to their
        __init__ methods.
        """
        super().__init__(config)

        # Configure the model and data.
        self.model = model_cls(config)
        self.data = data_module_cls(config)

        # Instantiate the callback classes for the trainer.
        if callback_classes:
            self.callbacks = [c_cls(config) for c_cls in callback_classes]
        else:
            self.callbacks = []

        # Instantiate the trainer.
        params = config.experiment.params
        trainer_kwargs = get_container(params, "trainer", default={}, resolve=True)
        trainer_kwargs["callbacks"] = self.callbacks
        self.trainer = lightning.Trainer(**trainer_kwargs)

        # Configure MLflow logging.
        # https://github.com/mlflow/mlflow/blob/eb3588abb55032838b1ff5af1b7d660f2b826dee/examples/pytorch/MNIST/mnist_autolog_example.py
        if self.trainer.global_rank == 0:
            mlflow.pytorch.autolog()

        # Dict for storing arbitrary result values.
        self.results = {}

    def get_objective_value(self) -> OptimizationValue:
        """
        Return an objective value for the optimizer. This must be overridden in
        subclasses. It can use any of the attributes of this class such as the
        LightningModule instance or any of the optional Callback instances.

        Returns:
            A numerical value or tuple of numerical values for the optimizer.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not override get_objective_value()"
        )

    def __call__(self):
        """
        Fit and test the model using the trainer and return the objective value.
        """
        self.trainer.fit(self.model, datamodule=self.data)
        self.results["test"] = self.trainer.test(datamodule=self.data)
        return self.get_objective_value()
